\ IR Detection Code
\ Written for AmForth
\ By: Craig A. Lindley
\ Last Update: 12/15/2013
\ Must have bitnames.frt and buffer.frt loaded

marker _IRDETECT_

\ Define port and pin used for IR detector
PORTD 4 portpin: IRDETECTOR

\ RAM buffer for storage of raw IR data
100    constant BUFSIZ
BUFSIZ buffer:  buffer

1500 constant MAX_COUNT

variable index
variable count

\ RAM buffer access functions
: buf!			( data index --- )
    buffer + c!
;

: buf@			( index --- data )
    buffer + c@
;

\ Clear the buffer
: clearBuffer
    BUFSIZ 0
    do
        0 i buf!
    loop
;

\ Read the IR transitions (MARKs and SPACEs) into the buffer
\ Recording edge detection times
: readRawIRCode

    clearBuffer

    0 index !
    begin
        0 count !

        \ Do we still have a SPACE ?
        begin
            IRDETECTOR pin_high?
        while
            count @ MAX_COUNT u>=
            if exit then
            1 count +!
            noop
        repeat

        \ No timout so store count
        count @ index @ buf!
        1 index +!

        0 count !

        \ Do we still have a MARK ?
        begin
            IRDETECTOR pin_low?
        while
            count @ MAX_COUNT u>=
            if exit then
            1 count +!
            noop
        repeat

        \ No timout so store count
        count @ index @ buf!
        1 index +!

        0
    until
;

\ Compare count values with tolerance
\ Returns 0 if new is shorter, 1 if equal and 2 if new is longer
\ A 20% tolerance is used in the compare

: compare		( old new --- 0 | 1 | 2 )
    over over 		( old new --- old new old new )
    swap		( old new old new --- old new new old )
    8 * 10 / <		( old new new old --- old new f )
    if			( old new f --- old new )
        drop drop 0
    else
        8 * 10 / <
        if
            2
        else
            1
        then
    then
;

\ 8 bit permutation table for hashing
create PERMUTTBL
     98 ,   6 ,  85 , 150 ,  36 ,  23 , 112 , 164 ,
    135 , 207 , 169 ,   5 ,  26 ,  64 , 165 , 219 , 
     61 ,  20 ,  68 ,  89 , 130 ,  63 ,  52 , 102 ,
     24 , 229 , 132 , 245 ,  80 , 216 , 195 , 115 ,
     90 , 168 , 156 , 203 , 177 , 120 ,   2 , 190 ,
    188 ,   7 , 100 , 185 , 174 , 243 , 162 ,  10 ,
    237 ,  18 , 253 , 225 ,   8 , 208 , 172 , 244 ,
    255 , 126 , 101 ,  79 , 145 , 235 , 228 , 121 ,
    123 , 251 ,  67 , 250 , 161 ,   0 , 107 ,  97 ,
    241 , 111 , 181 ,  82 , 249 ,  33 ,  69 ,  55 ,
     59 , 153 ,  29 ,   9 , 213 , 167 ,  84 ,  93 ,
     30 ,  46 ,  94 ,  75 , 151 , 114 ,  73 , 222 ,
    197 ,  96 , 210 ,  45 ,  16 , 227 , 248 , 202 ,
     51 , 152 , 252 , 125 ,  81 , 206 , 215 , 186 ,
     39 , 158 , 178 , 187 , 131 , 136 ,   1 ,  49 ,
     50 ,  17 , 141 ,  91 ,  47 , 129 ,  60 ,  99 ,
    154 ,  35 ,  86 , 171 , 105 ,  34 ,  38 , 200 ,
    147 ,  58 ,  77 , 118 , 173 , 246 ,  76 , 254 ,
    133 , 232 , 196 , 144 , 198 , 124 ,  53 ,   4 ,
    108 ,  74 , 223 , 234 , 134 , 230 , 157 , 139 ,
    189 , 205 , 199 , 128 , 176 ,  19 , 211 , 236 ,
    127 , 192 , 231 ,  70 , 233 ,  88 , 146 ,  44 ,
    183 , 201 ,  22 ,  83 ,  13 , 214 , 116 , 109 ,
    159 ,  32 ,  95 , 226 , 140 , 220 ,  57 ,  12 ,
    221 ,  31 , 209 , 182 , 143 ,  92 , 149 , 184 ,
    148 ,  62 , 113 ,  65 ,  37 ,  27 , 106 , 166 ,
      3 ,  14 , 204 ,  72 ,  21 ,  41 ,  56 ,  66 ,
     28 , 193 ,  40 , 217 ,  25 ,  54 , 179 , 117 ,
    238 ,  87 , 240 , 155 , 180 , 170 , 242 , 212 ,
    191 , 163 ,  78 , 218 , 137 , 194 , 175 , 110 ,
     43 , 119 , 224 ,  71 , 122 , 142 ,  42 , 160 ,
    104 ,  48 , 247 , 103 ,  15 ,  11 , 138 , 239 ,

\ Permutation table access function
: permut@		( index --- data )
    PERMUTTBL + @i
;

variable hashCode

\ Walk through the counts in the buffer comparing space and mark lengths
\ Do an 8 bit Pearson hash of the result to return as key code
: decodeIRCode		( --- keycode )

    0 hashCode !
    index @ 1		\ Ignore the first space code
    do
        i buf@ i 2 + buf@
        compare

        hashCode @ xor
        permut@ hashCode !
    loop
    
    \ Return result
    hashCode @
;

\ See if an IR code is available
: iRCodeAvailable	( --- f )
    IRDETECTOR pin_input
    -int
    readRawIRCode
    +int
    index @
;

\ Wait until an IR key code is available
: waitForIRCode		( --- key )
    IRDETECTOR pin_input
    begin
        -int
        readRawIRCode
        +int

        \ Index will be non zero when a key is detected
        index @ 
    until
    decodeIRCode
;

\ Adafruit Remote Control Codes
\ NOTE: SETUP and DOWNARROW hash the same. Use context to distinguish

&93  constant IRC_VOL-
&125 constant IRC_PLAYPAUSE
&132 constant IRC_VOL+
&88  constant IRC_SETUP
&40  constant IRC_UPARROW
&111 constant IRC_STOPMODE
&14  constant IRC_LEFTARROW
&121 constant IRC_ENTERSAVE
&194 constant IRC_RIGHTARROW
&88  constant IRC_DOWNARROW
&73  constant IRC_BACK
&198 constant IRC_0KEY
&188 constant IRC_1KEY
&229 constant IRC_2KEY
&87  constant IRC_3KEY
&201 constant IRC_4KEY
&150 constant IRC_5KEY
&32  constant IRC_6KEY
&199 constant IRC_7KEY
&184 constant IRC_8KEY
&46  constant IRC_9KEY

\ Check for a specific key
: checkForKey		( key --- f )
    waitForIRCode
    =
;














