\ Code Module for Si-4703 FM Receiver
\ Written for AmForth
\ By: Craig A. Lindley
\ Last Update: 12/01/2013
\ Must have marker.frt, bitnames.frt, buffer.frt and twi.frt loaded

marker _SI4703_

\ Define some port pins
PORTC 5 portpin: SCLK
PORTC 4 portpin: SDIO
PORTD 2 portpin: RESET
PORTD 6 portpin: STEREOIND

\ The Si4703 TWI address
$20 constant SI4703

\ Important register names
&00 constant DEVICEID
&01 constant CHIPID
&02 constant POWERCFG
&03 constant CHANNEL
&04 constant SYSCONFIG1
&05 constant SYSCONFIG2
&07 constant TEST1
&10 constant STATUSRSSI
&11 constant READCHAN
&12 constant RDSA
&13 constant RDSB
&14 constant RDSC
&15 constant RDSD

\ Register 2 POWERCFG bit numbers
&08 constant SEEK
&09 constant SEEKUP
&10 constant SKMODE
&14 constant DMUTE
&15 constant DSMUTE

\ Register 3 CHANNEL bit numbers
&15 constant TUNE

\ Register 4 SYSCONFIG1 bit numbers
&12 constant RDS

\ Register 10 STATUSRSSI bit numbers
&08 constant STEREO
&13 constant SFBL
&14 constant STC
&15 constant RDSR

\ Create register order mapping table
create ROTBL
 9 , 8 , 7 , 6 , 5 , 4 , 3 , 2 , 1 , 0 , 15 , 14 , 13  , 12 , 11 , 10 ,

: regOrder@ ROTBL + @i ;

\ Create register array of 32 RAM bytes
32 buffer: REGS

\  Cell access functions
: reg@ cells REGS + @ ;		( index --- data )
: reg! cells REGS + ! ;		( data index --- )

\ Receive N bytes from address
: <i2c				( N addr --- )
    twi.start
    1+ twi.tx			\ Send reception address
    1- dup 0 > 
    if
        0 do 
            twi.rx
       loop
    else
        drop
    then
    twi.rxn
    twi.stop
; 

\ Display the reg array for debugging
: displayRegisters
    hex cr
    16 0
    do
        i u. space i reg@ u. cr
    loop
    decimal cr
;

\ Read the register and assemble them into the correct order
: readRegisters			(  ---  )
   32 SI4703 <i2c		\ 32 bytes of register data on stack
   16 0
   do				\ Grab bytes from stack
       swap
       8 lshift or		\ Convert to word
       i regOrder@ reg!		\ Store in appropriate register
   loop
;

\ Update the Si4703 registers
: updateRegisters
    twi.start
    SI4703 twi.tx		\ Write the address
    TEST1 1+ POWERCFG		\ Write control registers 2 thru 7 
    do
        i reg@ dup
        8 rshift twi.tx		\ Write high byte
        $FF and  twi.tx		\ Write low byte
    loop
    twi.stop
;

\ Initialize the Si4703 hardware
: initSI4703
    
    \ Set port pins to output
    RESET pin_output
    SDIO  pin_output
    SCLK  pin_output
    STEREOIND pin_output

    STEREOIND low		\ Turn off the stereo indicator

    \ Set the Si 4703 into 2 wire mode via reset
    10 ms
    SDIO  low
    SCLK  high
    RESET low
    10 ms
    RESET high
    10 ms   

    \ Setup TWI to run at 333KHz on 5V 16000000Hz Arduino Uno
    1 4 twi.init

    readRegisters		\ Read the 16 registers into RAM
    
    $8100 TEST1 reg!		\ Enable the oscillator
    
    updateRegisters		\ Update registers
    
    500 ms			\ Wait for clock to settle

    readRegisters

    $4001 POWERCFG reg!		\ Enable Si4703
    
    SYSCONFIG1 reg@
    1 RDS lshift or		\ Enable RDS
    SYSCONFIG1 reg!

    SYSCONFIG2 reg@
    $FFC0 and			\ US channel spacing 
    SYSCONFIG2 reg!		\ Volume to off
 
    updateRegisters

    120 ms			\ More than max power up time
;

\ Turn off or close the receiver
: close				( --- )
    readRegisters
    $0000 POWERCFG reg!
    updateRegisters
;

\ Set the volume in range 0 .. 15
: setVolume			( volume --- )
    >r
    r@ 0 <
    if
        r> drop 0 >r
    then
    r@ 15 >
    if
        r> drop 15 >r
    then
    readRegisters
    SYSCONFIG2 reg@
    $FFF0 and			\ Clear volume bits
    r> or			\ Set new volume bits
    SYSCONFIG2 reg!
    updateRegisters
;

\ Muting control true is mute; false is unmute
: muteControl			( f --- )
    readRegisters
    POWERCFG reg@
    swap
    if				\ Muting
        1 DMUTE lshift invert and
    else			\ Not muting
        1 DMUTE lshift or
    then
    POWERCFG reg!
    updateRegisters
;

\ Soft muting control true is mute; false is unmute
: softMuteControl			( f --- )
    readRegisters
    POWERCFG reg@
    swap
    if				\ Muting
        1 DSMUTE lshift invert and
    else			\ Not muting
        1 DSMUTE lshift or
    then
    POWERCFG reg!
    updateRegisters
;

\ Set the FM channel. 97.3 input as 973
: setChannel			( channel --- )

    10 *			\ See AN230
    8750 -
    10 /
    			
    readRegisters
    CHANNEL reg@
    $FE00 and			\ Clear channel bits
    or				\ Set new channel bits
    1 TUNE lshift or		\ Set tune bit
    CHANNEL reg!
    updateRegisters
    
    \ Poll to see if STC is set
    begin
        readRegisters
        STATUSRSSI reg@		\ Read the status register
        1 STC lshift and
        0 <>
    until

    \ Clear the tune bit after tuning is complete
    readRegisters
    CHANNEL reg@
    1 TUNE lshift invert and
    CHANNEL reg!
    updateRegisters

    \ Wait for the STC bit to clear as well
    begin
        readRegisters
        STATUSRSSI reg@		\ Read the status register
        1 STC lshift and
        0=
    until
;

\ Get the currently set channel
: getChannel			( --- channel )

    readRegisters
    READCHAN reg@
    $03FF and			\ Isolate lower 10 bits
    875 +			\ Add FM freq offset
;

\ Seek to next station. True seeks up False seeks down
variable sfblValue

: seek				( f --- channel|0 )

    readRegisters
    POWERCFG reg@
    1 SKMODE lshift or swap	\ Allow seek to wrap
    if				\ Seek up
        1 SEEKUP lshift or	\ Set bit to seek higher
    else
        1 SEEKUP lshift invert and \ Clear bit to seek lower
    then
    1 SEEK lshift or		\ Start the seek process
    POWERCFG reg!  
    updateRegisters

    \ Poll to see if STC is set
    begin
        readRegisters
        STATUSRSSI reg@		\ Read the status register
        1 STC lshift and
        0 <>
    until

    readRegisters
    STATUSRSSI reg@
    1 SFBL lshift and
    sfblValue !			\ Save SFBL value

    POWERCFG reg@
    1 SEEK lshift invert and	\ Clear the seek bit
    POWERCFG reg!
    updateRegisters

    \ Wait for the STC bit to clear as well
    begin
        readRegisters
        STATUSRSSI reg@		\ Read the status register
        1 STC lshift and
        0=
    until

    \ Was a channel found ?
    sfblValue @ 0=
    if
        getChannel		\ Yes return the channel found			    
    else
        0			\ No channel was not found return 0
    then
;

\ Update the state of the stereo indicator
: updateReceptionIndicator	( ---  ) 

    readRegisters
    STATUSRSSI reg@
    1 STEREO lshift and
    0 <>
    if
        STEREOIND high
    else
        STEREOIND low
    then
;

: setup initSI4703 909 setChannel 1 setVolume ;


